import unittest
import time
import random

from scullery import workers


def workerTest(extra=0):
    l = []

    def f():
        l.append(1)

    workers.do(f)
    for i in range(200000):
        if l:
            break
        time.sleep(0.00001)
    time.sleep(extra)
    return len(l)


def workerSpam(count=10):
    l = []

    def f():
        time.sleep(random.random() * 0.05)
        l.append(1)

    for i in range(count):
        workers.do(f)


assert workerTest()


class TestWorkers(unittest.TestCase):
    def test_Workers(self):
        workers.start(8)

        workerSpam(100)
        self.assertLessEqual(len(workers.workers), workers.maxWorkers)
        self.assertGreaterEqual(len(workers.workers), workers.minWorkers + 2)

        for i in range(10):
            self.assertEqual(workerTest(), 1)

        for i in range(10):
            self.assertEqual(workerTest(), 1)
            time.sleep(0.001)

        for i in range(1):
            self.assertEqual(workerTest(0.1), 1)

        for i in range(100):
            self.assertEqual(workerTest(), 1)
            time.sleep(random.random() * 0.001)

        self.assertLessEqual(len(workers.workers), workers.maxWorkers)

        time.sleep(3)
        self.assertGreaterEqual(len(workers.workers), workers.minWorkers)

        for i in range(25):
            if workers.workers:
                time.sleep(1)
        # Workers should stop within 3s of inactivity, but we can only stop one at a time
        # every seconds
        self.assertEqual(len(workers.workers), 0)



from scullery import messagebus
import unittest
import gc
import time


class TestMsgbus(unittest.TestCase):
    def test_sync(self):
        p = []

        def f(t, v):
            p.append((t, v))

        messagebus.subscribe("/test", f)
        messagebus.post_message("/test", "foo", synchronous=True)

        self.assertEqual(p, [("/test", "foo")])

    def test_del(self):
        "Test unsubscribing by deleting a function"
        p = []

        def f(v):
            p.append(v)

        messagebus.subscribe("test", f)
        del f
        gc.collect()

        messagebus.post_message("test", "foo", synchronous=True)

        self.assertEqual(p, [])

    def test_async(self):
        p = []

        def f(v):
            p.append(v)

        messagebus.subscribe("test", f)
        messagebus.post_message("test", "foo")

        s = 100
        while (not p) and s:
            time.sleep(0.01)
            s -= 1
        self.assertEqual(p, ["foo"])



"""
Test Setup file
"""

import logging
import sys
import pytest


@pytest.fixture(scope="function", autouse=True)
def test_log(request):
    print(
        "Test '{}' STARTED".format(request.node.nodeid)
    )  # Here logging is used, you can use whatever you want to use for logs

    def fin():
        print("Test '{}' COMPLETED".format(request.node.nodeid))

    request.addfinalizer(fin)



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

import stat
import io
import sys
import os
import weakref
import threading
import gzip
import bz2
import json
import time
import logging
import hashlib
import threading

posix_rename = False
if sys.platform.startswith('linux'):
    posix_rename = True
if sys.platform.startswith('darwin'):
    posix_rename = True

# Purely just a dict for the rest of the application to keep track of what files are changed and not saved.
unsavedFiles = {}


def resolve_path(fn, expand=True):
    if not expand:
        return fn
    return (os.path.expandvars(os.path.expanduser(fn)))


# TODO: Ensure only one thread can save a file at a time
strio = io.BytesIO

persisters = []

# must be Rlock because load is recursive
lock = threading.RLock()


def saveExecutor(f):
    "Replace this with something that can call f with no args"
    f()


def saveAllAtExit():
    while persisters:
        i = persisters.pop()()
        try:
            i.save()
        except Exception:
            logging.exception("err")


def chmod_private_try(p, execute=True):
    try:
        if execute:
            os.chmod(p, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR |
                     stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP)
        else:
            os.chmod(p, stat.S_IRUSR | stat.S_IWUSR |
                     stat.S_IRGRP | stat.S_IWGRP)
    except Exception as e:
        raise e


def ensure_dir(f):
    d = os.path.dirname(f)
    if not os.path.exists(d):
        os.makedirs(d)


class Persister():
    def __init__(self, fn, default=None):
        self.fn = fn
        try:
            self.reload()
        except Exception:
            self.value = default
        torm = []
        # Before we add ourselves, clear out any old persisters that are no longer needed.
        for i in persisters:
            if not i():
                torm.append(i)
        for i in torm:
            persisters.remove(i)

        persisters.append(weakref.ref(self))

    def save(self):
        save(self.value, self.fn, backup=True)

    def reload(self):
        if os.path.exists(self.fn):
            self.value = load(self.fn)


def save(data, fn, *, private=False, backup=True, expand=True, md5=False, nolog=False):
    """Save data to file. Filename must end in .json, .yaml, .txt, or .bin. Data will be encoded appropriately.
        Also supports compressed versions via filenames ending in .gz or .bz2.
        Args:
            data:
                the data to be written. if fn is a .json or .yaml, must be serializable. If filename is .txt, must be a string.
                If .bin, must be something like bytes.
            private:
                If True, file created with mode 700(Full access to root and owner but not even read to anyone else)
                If False(the default), file created with default mode
            backup:
                Setting this to true is an alias for mode="backup"
    """
    fn = resolve_path(fn, expand)
    with lock:

        # Make sure we don't overwrite a file when we create our dirs, because that behavior is undocumented in makedirs.
        x = os.path.split(fn)[0]
        already = {}

        # Safety counter to stop really wierd loops
        for i in range(64):
            if os.path.isfile(x):
                raise RuntimeError(
                    "Required intermediate directory is already present as a file, refusing to overwrite file")
            x = os.path.split(x)[0]
            # Loop prevention
            if x in already:
                break
            already[x] = True

        if not os.path.exists(os.path.dirname(fn)):
            if private:
                os.makedirs(os.path.dirname(fn), mode=0o700)
            else:
                os.makedirs(os.path.dirname(fn))

        if os.path.isdir(fn):
            raise RuntimeError(
                "Filename is already present as a directory, refusing to overwrite directory")
        # Get base type without compression
        if fn.endswith(".gz"):
            x = fn[:-3]
        elif fn.endswith(".bz2"):
            x = fn[:-4]
        else:
            x = fn
        # Encode the data into our chosen format
        if x.endswith(".json"):
            data = json.dumps(data).encode('utf8')
        elif x.endswith(".yaml"):
            import yaml
            data = yaml.dump(data).encode('utf8')

        elif x.endswith(".toml"):
            import toml
            data = toml.dumps(data).encode("utf8")

        elif x.endswith(".txt") or x.endswith(".md") or x.endswith(".rst"):
            data = (str(data).encode('utf8'))
        elif x.endswith(".bin"):
            data = (data)
        else:
            raise ValueError('Unsupported or missing File Extension')

        # We have selected a compressed type. Compress in-memory first so we can read-before-write
        # Note that disk access is slow enough the call to  basically makes no difference in speed here if it's already imported
        if fn.endswith(".gz"):
            i = strio()
            f = gzip.GzipFile(fn, mode='wb', fileobj=i)
            f.write(data)
            f.close()
            data = i.getvalue()

        elif fn.endswith(".bz2"):
            c = bz2.BZ2Compressor()
            c.compress(data)
            data = c.flush()
            del c

        # Do a read-before-write. We don't write if we don't have to
        if os.path.exists(fn):
            with open(fn, "rb") as f:
                if f.read() == data:
                    return

        ensure_dir(os.path.split(fn)[0])

        if backup:
            tempfn = fn+str(time.time())
        else:
            tempfn = fn

        if not nolog:
            logging.debug("Writing: "+fn)
        # Actually write it
        with open(tempfn, 'wb') as f:

            # In backup mode, pre truncate and flush.
            # This means that even if an error occors during writing, we will be able to tell by the
            # mtime that the tilde file is the right one.
            if backup:
                f.flush()
                os.fsync(f.fileno())

            # Chmod it before we write anything to it.
            if private:
                chmod_private_try(tempfn)
            f.write(data)
            f.flush()
            os.fsync(f.fileno())

        if backup:
            os.replace(tempfn, fn)

        if md5:
            with open(fn + ".md5", "w") as md5f:
                md5f.write(hashlib.md5(data).hexdigest())


def load(filename, *, expand=True):
    """Load a file. Return str if file extension is .txt, bytes on .bin, dict on .yaml or .json.

    After that may be a .bz2 or a .gz for compression.

    """
    filename = resolve_path(filename, expand)

    with lock:
        try:
            # Open the file and get the filename without the compression type attached to it.
            if filename.endswith(".gz"):
                f = gzip.GzipFile(filename, mode='rb')
                x = filename[:-3]
            elif filename.endswith(".bz2"):
                x = filename[:-4]
                f = bz2.BZ2File(filename, mode='rb')
            else:
                f = open(filename, 'rb')
                x = filename

            if x.endswith(".json"):
                r = json.loads(f.read().decode('utf8'))
            elif x.endswith(".yaml"):
                import yaml
                r = yaml.load(f.read().decode('utf8'), Loader=yaml.SafeLoader)

            elif x.endswith(".toml"):
                import toml
                r = toml.loads(f.read().decode('utf8'))

            elif x.endswith(".txt") or x.endswith(".md") or x.endswith(".rst"):
                r = f.read().decode('utf8')
            elif x.endswith(".bin"):
                r = f.read()
            else:
                raise ValueError('Unsupported File Extension')
        except Exception as e:
            try:
                f.close()
            except Exception:
                pass
            raise
        try:
            f.close()
        except Exception:
            pass

        return r



import unittest
from scullery import mnemonics
from scullery import units


class TestMisc(unittest.TestCase):
    def test_wordlist(self):
        self.assertEqual(len(mnemonics.wordlist), 1633)

    def test_hash(self):
        "This just happens to be what you get if you hash that string"
        self.assertEqual(mnemonics.memorableBlakeHash(b"", 3, ""), "initialpuzzlegroup")

    def test_hash8(self):
        "Test with a longer requested result"
        self.assertEqual(
            mnemonics.memorableBlakeHash(b"", 8, ""),
            "initialpuzzlegroupdeclarefactoragendalorenzoacademy",
        )

    def test_conversions(self):
        "Test with a longer requested result"
        self.assertEqual(units.convert(12, "in", "ft"), 1)
        self.assertEqual(units.convert(1, "km", "m"), 1000)
        self.assertAlmostEqual(units.convert(50, "degC", "degF"), 122)
        self.assertAlmostEqual(units.convert(122, "degF", "degC"), 50)
        self.assertAlmostEqual(units.convert(1, "kft", "in"), 12000)
        # Milli-inches?
        self.assertAlmostEqual(units.convert(1, "ft", "min"), 12000)

    def test_nonsense_convert(self):
        with self.assertRaises(Exception):
            units.convert("80", "trashpiles", "garbages")
        # Can't use SI prefixes with nonlinear units like degrees
        with self.assertRaises(Exception):
            units.convert("80", "mdegC", "K")



import unittest
import time

from scullery import scheduling


class TestScheduler(unittest.TestCase):
    def test_scheduler(self):
        c = [0]

        @scheduling.every(1)
        def f():
            c[0] += 1

        time.sleep(1.5)
        assert c[0] > 0
        assert c[0] < 3

        time.sleep(1.5)
        assert c[0] > 1
        assert c[0] < 4

        f.unregister()
        x = c[0]

        time.sleep(2)

        assert x == c[0]



import time
from scullery import statemachines


class Obj:
    def __init__(self) -> None:
        # We turn this var on and off
        # to make sure we can
        self.sm_lamp = 0


def test_state_machines():
    blink = statemachines.StateMachine(start="off")

    blink.add_state("on")
    blink.add_state("off")

    blink.set_timer("on", 1, "off")

    blink.set_timer("off", 1, "on")
    blink.add_rule("off", "begin", "on")
    blink("begin")

    container = Obj()

    container.sm_lamp = 0

    def on():
        container.sm_lamp = 1

    def off():
        container.sm_lamp = 0

    sm = statemachines.StateMachine(start="off")

    sm.add_state("on", enter=on)
    sm.add_state("off", enter=off)

    sm.set_timer("on", 1, "off")

    sm.add_rule("off", "motion", "on")

    if container.sm_lamp:
        raise RuntimeError("state machine imaginary lamp is on too soon")
    sm.event("motion")
    time.sleep(0.3)
    if not container.sm_lamp:
        raise RuntimeError("state machine imaginary lamp is not on")
    time.sleep(2)

    if container.sm_lamp:
        time.sleep(8)
        if container.sm_lamp:
            raise RuntimeError(
                "state machine imaginary lamp didn't turn itself off within 10s"
            )
        else:
            raise RuntimeError(
                "state machine imaginary lamp didn't turn itself off within 2s"
            )

    # Turn it on before deleting so we can make sure the timer won't trigger after it's gone
    sm.event("motion")
    del sm

    sm = statemachines.StateMachine(start="off")

    sm.add_state("on", enter=on)
    sm.add_state("off", enter=off)

    lights_on = [False]

    last_poll = [0.0]

    # Polled function trigger, when f returns true
    # We should go to the next state
    def f():
        last_poll[0] = time.time()
        return lights_on[0]

    sm.add_rule("on", f, "off")
    sm.add_rule("off", "motion", "on")

    if not sm.state == "off":
        raise RuntimeError("Unexpected state")

    sm.event("motion")
    if not sm.state == "on":
        raise RuntimeError("Unexpected state")

    # Make sure polling doesn't trigger it if false
    time.sleep(0.3)
    if not sm.state == "on":
        raise RuntimeError("Unexpected state")

    lights_on[0] = True
    time.sleep(0.3)

    if not sm.state == "off":
        raise RuntimeError("Unexpected state")

    t = last_poll[0]
    if not t:
        raise RuntimeError("Test probably is buggy")

    time.sleep(0.3)
    if not last_poll[0] == t:
        raise RuntimeError("State machine continued polling after exiting event")

    lights_on[0] = False
    time.sleep(0.25)
    t = last_poll[0]
    del sm
    time.sleep(1)

    if not last_poll[0] == t:
        raise RuntimeError("State machine continued polling after deleting")

    transitions = []

    sm = statemachines.StateMachine(start="off")

    sm.add_state("on", enter=on)
    sm.add_state("off", enter=off)

    sm.add_rule("off", "switch", "on")

    def subscriber(s):
        transitions.append(s)

    sm.subscribe(subscriber)
    sm.event("switch")
    if not transitions == ["on"]:
        print(transitions)
        raise RuntimeError("State machines not working as they should")



import time
import scullery.ratelimits


def test_rate_limits():
    # 1hz average, up to max burst of 50
    rl = scullery.ratelimits.RateLimiter(hz=1, burst=50)
    found_limit = 0

    for i in range(100):
        # Returns number of credits remaining.
        # They refill at the given hz rate up to the burst limit

        if not rl.limit():
            found_limit = i
            break

    # Time based so allow some slop, but generally confirm
    # it allowed a 50 round burst and stopped
    assert abs(found_limit - 50) < 5

    rl = scullery.ratelimits.RateLimiter(hz=10, burst=3)
    found_limit = 0

    for i in range(10):
        time.sleep(0.1)
        rl.limit()

    assert abs(rl.limit() - 3) < 2
